// created on 6/6/2004 at 10:46 AM
/*
 *   Copyright (c) 2004, Alexandros Frantzis (alf82 [at] freemail [dot] gr)
 *
 *   This file is part of Bless.
 *
 *   Bless is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Bless is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Bless; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

using System.Collections;
using System;
using System.IO;
using System.Threading;
using Bless.Util;

namespace Bless.Buffers {

///<summary>
/// A buffer for holding bytes in a versatile manner.
/// It supports undo-redo and can easily handle large files.
/// Editing is also very cheap.
///</summary>
public class ByteBuffer {

	internal ChangeBuffer changeBuf;
	internal FileBuffer	fileBuf;
	internal SegmentCollection segCol;
	Stack undoStack;
	Stack redoStack;
	object SaveCheckpoint;
	FileSystemWatcher fsw;
	
	// automatic file naming
	string autoFilename;
	static int autoNum=1;
	
	internal long size;
	
	// chaining related
	bool actionChaining;
	bool actionChainingFirst;
	
	// buffer permissions
	readonly public object LockObj=new object();
	bool readAllowed;
	bool modifyAllowed;
	bool fileOperationsAllowed;
	bool emitEvents;
	
	public delegate void ChangedHandler(ByteBuffer bb);
	
	///<summary>Emitted when buffer changes</summary>
	public event ChangedHandler Changed;
	///<summary>Emitted when file changes (outside DataView)</summary>
	public event ChangedHandler FileChanged;
	
	// related to asynchronous save model
	AsyncCallback userSaveAsyncCallback;
	AsyncCallback userSaveAsAsyncCallback;
	AutoResetEvent saveAsFinishedEvent;
	AutoResetEvent saveFinishedEvent;
	bool useGLibIdle;
	
	public ByteBuffer() 
	{
		segCol=new SegmentCollection();
		changeBuf= new ChangeBuffer();
		undoStack= new Stack();
		redoStack= new Stack();
		size=0;
		SaveCheckpoint=null;
		
		// name the buffer automatically
		autoFilename="Untitled " + ByteBuffer.autoNum;
		ByteBuffer.autoNum++;
		readAllowed=true;
		fileOperationsAllowed=true;
		modifyAllowed=true;
		saveFinishedEvent=new AutoResetEvent(false);
		saveAsFinishedEvent=new AutoResetEvent(false);
		useGLibIdle=false;
		emitEvents=true;
	}
	
	///<summary>Create a ByteBuffer loaded with a file</summary>
	static public ByteBuffer FromFile(string filename)
	{
		ByteBuffer bb=new ByteBuffer();
		bb.LoadWithFileBuffer(new FileBuffer(filename, 0xfffff));
		
		// fix automatic file naming
		ByteBuffer.autoNum--;
		
		return bb;
	}
	
	///<summary>Create a ByteBuffer with a dummy-name</summary>
	public ByteBuffer(string filename): this() 
	{
		this.autoFilename=filename;
		
		// fix automatic file naming
		ByteBuffer.autoNum--;
	}
	
	///<summary>Regard all following actions as a single one</summary>
	public void BeginActionChaining()
	{
		actionChaining=true;
		actionChainingFirst=true;
		emitEvents=false;
	}
	
	///<summary>Stop regarding actions as a single one</summary>
	public void EndActionChaining()
	{
		actionChaining=false;
		actionChainingFirst=false;
		emitEvents=true;
		
		if (undoStack.Count > 0) {
			ByteBufferAction bba=(ByteBufferAction)undoStack.Peek();
			bba.ChainedNext=false;
		}
		
		if (Changed!=null)
			Changed(this);
	}
	
	///<summary>Set the chaining information of a ByteBufferAction</summary>
	void SetChaining(ByteBufferAction bba)
	{
		if (actionChainingFirst==true) {
			actionChainingFirst=false;
			bba.ChainedNext=true;
		}
		else if (actionChaining) {
			bba.ChainedPrevious=true;
			bba.ChainedNext=true;
		}
	}
	
	///<summary>Append bytes at the end of the buffer</summary>
	public void Append(byte[] data) 
	{	
		lock (LockObj) {
			if (!modifyAllowed) return;
		
			AppendAction aa=new AppendAction(data, this);	
			aa.Do();
		
			SetChaining(aa);
			undoStack.Push(aa);
			redoStack.Clear();
			if (emitEvents && Changed!=null)
				Changed(this);
		}
	}

	///<summary>Insert bytes into the buffer</summary>
	public void Insert(long pos, byte[] data) 
	{
		lock (LockObj) {
			if (!modifyAllowed) return;
			
			if (pos == size) {
				Append(data);
				return;
			}
			InsertAction ia=new InsertAction(pos, data, this);
			ia.Do();
			
			SetChaining(ia);		
			undoStack.Push(ia);
			redoStack.Clear();
			if (emitEvents && Changed!=null)
				Changed(this);
		}
	}		

	///<summary>Delete bytes from the buffer</summary>
	public void Delete(long pos1, long pos2) 
	{
		lock (LockObj) {
			if (!modifyAllowed) return;
		
			DeleteAction da=new DeleteAction(pos1, pos2, this);
			da.Do();
		
			SetChaining(da);		
			undoStack.Push(da);
			redoStack.Clear();
			if (emitEvents && Changed!=null)
				Changed(this);
		}
	}
	
	///<summary>Replace bytes in the buffer</summary>
	public void Replace(long pos1, long pos2, byte[] data) 
	{
		lock (LockObj) {
			if (!modifyAllowed) return;
			
			ReplaceAction ra=new ReplaceAction(pos1, pos2, data, this);
			ra.Do();
		
			SetChaining(ra);
			undoStack.Push(ra);
			redoStack.Clear();
			if (emitEvents && Changed!=null)
				Changed(this);
		}
	}
	
	///<summary>Undo the last action or chain of actions</summary>
	public void Undo() 
	{
		lock (LockObj) {	
			if (!modifyAllowed) return;
		
			// while there are more actions
			while (undoStack.Count>0) {
				ByteBufferAction action=(ByteBufferAction)undoStack.Pop();
				action.Undo();
				redoStack.Push(action);
				
				// if this is the last action in the undo chain,
				// end the undo
				if (!action.ChainedPrevious) {
					if (emitEvents && Changed!=null)
						Changed(this);
					break;
				}
			}
		}
	}	
	
	///<summary>Redo the last undone action or chain of actions</summary>
	public void Redo() 
	{
		lock (LockObj) {	
			if (!modifyAllowed) return;
			
			// while there are more actions
			while (redoStack.Count>0) {
				ByteBufferAction action=(ByteBufferAction)redoStack.Pop();
				action.Do();
				undoStack.Push(action);
				
				// if this is the last action in the redo chain,
				// end the redo (note the use of ChainedNext and 
				// not ChainedPrevious, as in Undo())
				if (!action.ChainedNext) {
					if (emitEvents && Changed!=null)
						Changed(this);
					break;
				}
			}
		}
	}
	
	///<summary>
	/// Save the buffer as a file, using an asynchronous model
	///</summary>
	public IAsyncResult BeginSaveAs(string filename, ProgressCallback progressCallback, AsyncCallback ac)
	{
		lock (LockObj) {
			if (!fileOperationsAllowed) return null;
			
			saveAsFinishedEvent.Reset();
			userSaveAsAsyncCallback=ac;
			
			SaveAsOperation so=new SaveAsOperation(this, filename, progressCallback, SaveAsAsyncCallback, useGLibIdle);
		
			// don't allow messing up with the buffer
			// while we are saving
			this.ReadAllowed=false;
			this.ModifyAllowed=false;
			this.FileOperationsAllowed=false;
			this.EmitEvents=false;
			fsw.EnableRaisingEvents=false;
			
			// start save thread
			Thread saveThread=new Thread(so.OperationThread);
			saveThread.IsBackground=true;
			saveThread.Start();
			
			return new ThreadedAsyncResult(so, saveAsFinishedEvent, false);
		}
		
	}
	
	///<summary>
	/// Called when an asynchronous Save As operation finishes
	///</summary>
	void SaveAsAsyncCallback(IAsyncResult ar)
	{
		lock (LockObj) {
			SaveAsOperation so=(SaveAsOperation)ar.AsyncState;
			
			// re-allow buffer usage
			this.ReadAllowed=true;
			this.ModifyAllowed=true;
			this.FileOperationsAllowed=true;
			this.EmitEvents=true;
			
			
			// make sure Save As went smoothly before doing anything
			if (so.Result==SaveAsOperation.OperationResult.Finished) {
				CloseFile();
				LoadWithFileBuffer(new FileBuffer(so.SavePath));
				
				if (undoStack.Count > 0)
					SaveCheckpoint=undoStack.Peek();
				else
					SaveCheckpoint=null;
			}
			else { // if cancelled or caught an exception
				try {
					System.IO.File.Delete(so.SavePath);
				}
				catch(Exception ex) { }
			}
			
			fsw.EnableRaisingEvents=true;
			
			// if user provided a callback, call it now
			if (userSaveAsAsyncCallback!=null)
				userSaveAsAsyncCallback(ar);
			
			// notify that Save As has finished
			saveAsFinishedEvent.Set();
		}
	}
	
	///<summary>
	/// Save the buffer under the same filename, using an asynchronous model
	///</summary>
	public IAsyncResult BeginSave(ProgressCallback progressCallback, AsyncCallback ac) 
	{
		lock (LockObj) {		
			if (!fileOperationsAllowed) return null;
			
			saveFinishedEvent.Reset();
			userSaveAsyncCallback=ac;
			
			SaveOperation so=new SaveOperation(this, progressCallback, SaveAsyncCallback, useGLibIdle);
		
			// don't allow messing up with the buffer
			// while we are saving
			this.ReadAllowed=false;
			this.ModifyAllowed=false;
			this.FileOperationsAllowed=false;
			this.EmitEvents=false;
			fsw.EnableRaisingEvents=false;
			
			// start save thread
			Thread saveThread=new Thread(so.OperationThread);
			saveThread.IsBackground=true;
			saveThread.Start();
			
			return new ThreadedAsyncResult(so, saveFinishedEvent, false);
		}
	}
	
	///<summary>
	/// Called when an asynchronous save operation finishes
	///</summary>
	void SaveAsyncCallback(IAsyncResult ar)
	{
		lock (LockObj) {
			SaveOperation so=(SaveOperation)ar.AsyncState;
		
			// re-allow buffer usage
			this.ReadAllowed=true;
			this.ModifyAllowed=true;
			this.FileOperationsAllowed=true;
			this.EmitEvents=true;
			
			if (so.Result==SaveOperation.OperationResult.Finished) { // save went ok
				LoadWithFileBuffer(new FileBuffer(so.SavePath));
				
				if (undoStack.Count > 0)
					SaveCheckpoint=undoStack.Peek();
				else
					SaveCheckpoint=null;
			}
			else if (so.Result==SaveOperation.OperationResult.Cancelled) { // save cancelled
				if (so.StageReached==SaveOperation.SaveStage.BeforeSaveAs) {
					System.IO.File.Delete(so.TempPath);
				}
				else if (so.StageReached==SaveOperation.SaveStage.BeforeDelete) {
					System.IO.File.Delete(so.TempPath);
					fileBuf.Load(so.SavePath);
				}
				else if (so.StageReached==SaveOperation.SaveStage.BeforeMove) {
					// cancel has no effect during move.
					// mark operation as successful
					so.Result=SaveOperation.OperationResult.Finished;
					LoadWithFileBuffer(new FileBuffer(so.SavePath));
				
					if (undoStack.Count > 0)
						SaveCheckpoint=undoStack.Peek();
					else
						SaveCheckpoint=null;
				}
			}
			else if (so.Result==SaveOperation.OperationResult.CaughtException) {
				if (so.StageReached==SaveOperation.SaveStage.BeforeSaveAs) {
					System.IO.File.Delete(so.TempPath);
				}
				else if (so.StageReached==SaveOperation.SaveStage.BeforeDelete) {
					System.IO.File.Delete(so.TempPath);
					fileBuf.Load(so.SavePath);
				}
				else if (so.StageReached==SaveOperation.SaveStage.BeforeMove) {
					// TO-DO: better handling?
					fileBuf.Load(so.SavePath);
				}
			}
			
			fsw.EnableRaisingEvents=true;
			
			// if user provided a callback, call it now
			if (userSaveAsyncCallback!=null)
				userSaveAsyncCallback(ar);
			
			// notify that Save has finished	
			saveFinishedEvent.Set();
		}
	}
	
	///<summary> 
	/// Revert ByteBuffer to the last saved state
	///</summary> 
	public void Revert()
	{
		lock (LockObj) {
			if (!modifyAllowed) return;
			
			if (this.HasFile) {
				// reload file
				string filename=fileBuf.Filename;
				if (!File.Exists(filename))
					throw new FileNotFoundException(filename);
			
				fileBuf.Close();
				LoadWithFileBuffer(new FileBuffer(filename));

				undoStack.Clear();
				redoStack.Clear();
				SaveCheckpoint=null;
				
				// emit bytebuffer changed event
				if (emitEvents && Changed!=null)
					Changed(this);		
			}
		}
	}
	
	///<summary> 
	/// Sets the file buffer and resets the segment collection
	///</summary> 
	private void LoadWithFileBuffer(FileBuffer fb)
	{
		fileBuf=fb;
		Segment s=new Segment(fileBuf, 0, fileBuf.Size-1);
		segCol=new SegmentCollection();
		segCol.Append(s);
		size=fileBuf.Size;
		
		// monitor the file for changes
		//
		
		if (fsw!=null) {
			fsw.Dispose();
			fsw=null;
		}
			
		fsw=new FileSystemWatcher();
		fsw.Path=Path.GetDirectoryName(fb.Filename);
		fsw.Filter=Path.GetFileName(fb.Filename);
		fsw.NotifyFilter=NotifyFilters.FileName|NotifyFilters.LastAccess|NotifyFilters.LastWrite;
		fsw.Changed += new FileSystemEventHandler(OnFileChanged);
		//fsw.Deleted += new FileSystemEventHandler(OnFileChanged);
		
		fsw.EnableRaisingEvents = true;
	}
	
	private void OnFileChanged(object source, FileSystemEventArgs e)
	{
		if (emitEvents && FileChanged!=null)
			FileChanged(this);
	}
	
	public byte this[long index] {
		set { } 
		get {
			lock (LockObj) {
				long map; 
				Util.List.Node node;
				Segment seg=segCol.FindSegment(index, out map, out node);
				//Console.WriteLine("Searching index {0} at {1}:{2}", index, map, seg);
				if (seg==null)
					throw new IndexOutOfRangeException(string.Format("ByteBuffer[{0}]",index));
				else {
					try {
						return seg.buffer[seg.start+index-map];	
					}
					catch(IndexOutOfRangeException e) {
						Console.WriteLine("Problem at index {0} at {1}:{2}", index, map, seg);
						throw;
					}
				}
			}
		}
	}
	
	
	public void CloseFile()
	{
		lock (LockObj) {
			// close the file buffer and dispose the file watcher
			if (fileBuf != null && fileOperationsAllowed) {
				fileBuf.Close();
				fsw.Dispose();
				fsw=null;
			}
		}
	} 
	
	public long Size {
		get { return size;}
	}

	public bool HasFile {
		get { return fileBuf != null; }
			
	}
	
	public string Filename {
		get {
			if (fileBuf != null)
				return fileBuf.Filename;
			else
				return this.autoFilename; 
			}	
	}

	public bool HasChanged {
		get {
			if (undoStack.Count>0)
				return (SaveCheckpoint != undoStack.Peek());
			else
				return SaveCheckpoint != null;
		}
	
	}
	
	public bool CanUndo {
		get { return (undoStack.Count>0); }
	}
	
	public bool CanRedo {
		get { return (redoStack.Count>0); }
	}
	
	public bool ActionChaining {
		get { return actionChaining; }
	}
	
	// Whether the ByteBuffer will emit events
	// (eg Changed event)
	public bool EmitEvents {
		get { return emitEvents; }
		set { 
			emitEvents=value; 
			if (emitEvents && Changed!=null)
				Changed(this);
		}
	}
	
	// whether buffer can be safely read
	// by user eg to display data in a DataView.
	// This is only a hint and doesn't modify
	// the of behavior of the buffer.
	public bool ReadAllowed { 
		get { return readAllowed; }
		set { 
			readAllowed=value;
			if (emitEvents && Changed!=null)
				Changed(this);
		}
	}
	
	// Whether buffer can be modified.
	// If it is false, all buffer actions
	// that can modify the buffer are 
	// rendered ineffective.
	public bool ModifyAllowed {
		get { return modifyAllowed;}
		set { 
			modifyAllowed=value; 
			if (emitEvents && Changed!=null)
				Changed(this);
		}
	}
	
	// Whether buffer can be saved, closed etc
	// If it is false, all save and close operations
	// are ignored.
	public bool FileOperationsAllowed {
		get { return fileOperationsAllowed;}
		set { 
			fileOperationsAllowed=value; 
			if (emitEvents && Changed!=null)
				Changed(this);
		}
	}
	
	// Use the GLib Idle handler for progress reporting.
	// Mandatory if progress reporting involves Gtk+ widgets.
	public bool UseGLibIdle {
		get { return useGLibIdle; }
		set { useGLibIdle=value; }
	}
	
	internal void Display(string s) 
	{
		Console.Write(s);
		segCol.List.Display();
	}

}

} // end namespace 