// project created on 6/3/2004 at 8:34 PM
/*
 *   Copyright (c) 2004, Alexandros Frantzis (alf82 [at] freemail [dot] gr)
 *
 *   This file is part of Bless.
 *
 *   Bless is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Bless is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Bless; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

using System;
using System.Collections;
using System.Threading;
using System.IO;
using Gtk;
using Glade;
using Bless.Buffers;
using Bless.Gui;
using Bless.Gui.Dialogs;
using Bless.Tools.Find;
using Bless.Tools;
using Bless.Util;

public class BlessMain
{
	[Glade.Widget] Gtk.HBox DataViewBox;
	[Glade.Widget] Gtk.Toolbar MainToolbar;
	[Glade.Widget] Gtk.Window MainWindow;
	[Glade.Widget] Gtk.VBox MainVBox;
	
	[Glade.Widget] Gtk.Paned StatusPaned;
	[Glade.Widget] Gtk.Statusbar OffsetStatusbar;
	[Glade.Widget] Gtk.Statusbar SelectionStatusbar;
	[Glade.Widget] Gtk.Statusbar MessageStatusbar;
	[Glade.Widget] Gtk.Statusbar OverwriteStatusbar;
	[Glade.Widget] Gtk.Statusbar DummyStatusbar;
	
	[Glade.Widget] Gtk.MenuItem FileMenuItem;
	[Glade.Widget] Gtk.MenuItem FileSaveMenuItem;
	[Glade.Widget] Gtk.MenuItem FileSaveAsMenuItem;
	[Glade.Widget] Gtk.MenuItem FileCloseMenuItem;
	[Glade.Widget] Gtk.MenuItem FileQuitMenuItem;
	[Glade.Widget] Gtk.MenuItem FileRevertMenuItem;
	[Glade.Widget] Gtk.MenuItem EditCutMenuItem;
	[Glade.Widget] Gtk.MenuItem EditCopyMenuItem;
	[Glade.Widget] Gtk.MenuItem EditPasteMenuItem;
	[Glade.Widget] Gtk.MenuItem EditDeleteMenuItem;
	[Glade.Widget] Gtk.MenuItem EditUndoMenuItem;
	[Glade.Widget] Gtk.MenuItem EditRedoMenuItem;
	[Glade.Widget] Gtk.MenuItem SearchFindNextMenuItem;
	[Glade.Widget] Gtk.MenuItem SearchFindPreviousMenuItem;
	[Glade.Widget] Gtk.MenuItem HelpContentsMenuItem;
	[Glade.Widget] Gtk.CheckMenuItem ViewStatusbarShowMenuItem;
	[Glade.Widget] Gtk.CheckMenuItem ViewStatusbarOffsetMenuItem;
	[Glade.Widget] Gtk.CheckMenuItem ViewStatusbarSelectionMenuItem;
	[Glade.Widget] Gtk.CheckMenuItem ViewStatusbarOverwriteMenuItem;
	[Glade.Widget] Gtk.CheckMenuItem ViewConversionTableMenuItem;
	

	DataBook dataBook;
	FindDialog findDialog;
	ReplaceDialog replaceDialog;
	ProgressDialog findProgressDialog;
	GotoOffsetDialog gotoOffsetDialog;
	ConversionTable conversionTable;
	
	IFinder finder;
	History history;
	
	Timer StatusMessageTimer;
	
	Hashtable bufferHasChanged;
	
	public static void Main (string[] args)
	{
		new BlessMain (args);
	}

	public BlessMain (string[] args) 
	{
		Application.Init();
		
		// load main window from glade XML
		Glade.XML gxml = new Glade.XML (null, "bless.glade", "MainWindow", null);
		gxml.Autoconnect (this);
		
		// set the application icon
		MainWindow.Icon=new Gdk.Pixbuf(FileResourcePath.GetPath("..","data","bless-48x48.png"));
		
		// create timer for status bar messages
		StatusMessageTimer= new Timer(new TimerCallback(ClearStatusMessage), null, System.Threading.Timeout.Infinite, 0);
		
		// create a hash table to hold the status of each buffer
		// in the DataBook
		bufferHasChanged=new Hashtable();
		
		// create the DataBook
		dataBook=new DataBook();
		dataBook.SwitchPage+= new SwitchPageHandler(OnSwitchPage);
		
		DataViewBox.PackStart(dataBook);
		
		conversionTable=new ConversionTable();
		MainVBox.PackStart(conversionTable, false, false, 0);
		MainVBox.ReorderChild(conversionTable, 2);
		
		// create the find progress and find dialogs (hidden)
		findProgressDialog=new ProgressDialog("");
		finder=new FileFinder(dataBook, findProgressDialog.Update);
		finder.Strategy=new BMFindStrategy();
		finder.FirstFind+= OnFirstFind;
		SearchFindNextMenuItem.Sensitive=false;
		SearchFindPreviousMenuItem.Sensitive=false;
		findDialog=new FindDialog(finder);
		
		// create the replace dialog (hidden)
		replaceDialog=new ReplaceDialog(finder);
		
		// create the go to offset dialog (hidden)
		gotoOffsetDialog=new GotoOffsetDialog(dataBook);
		
		string blessConfDir=Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), "bless");
		
		// make sure local configuration directory exists
		try {
			if (!Directory.Exists(blessConfDir)) {
				Directory.CreateDirectory(blessConfDir);
			}
		}
		catch (Exception ex){
			ErrorAlert ea=new ErrorAlert("Cannot create user configuration directory", ex.Message+"\n\nSome features of Bless may not work properly.", MainWindow);
			ea.Run();
			ea.Destroy();
		}
		
		// load recent file history
		history=new History(5);
		history.Changed += OnHistoryChanged;
		LoadHistory(Path.Combine(blessConfDir,"history.xml"));
		
		// if user specified files on the command line
		// try to load them
		if (args.Length>0) {
			LoadFiles(args);
		}
		else { // try to load previous session
			LoadSession(Path.Combine(blessConfDir,"last.session"));
		}
		
		// if nothing has been loaded, create a new file
		if (dataBook.NPages == 0) {
			ByteBuffer bb=new ByteBuffer();
			bb.UseGLibIdle=true;
			bufferHasChanged[bb]=true;
			bb.Changed+=OnBufferChanged;
			// create and setup a  DataView
			DataView dv=CreateDataView(bb);
	
			// append the DataView to the DataBook
			string filename=GetBufferPath(bb);
			dataBook.AppendView(dv, new CloseViewDelegate(CloseFile), Path.GetFileName(filename));
	
			OnBufferChanged(bb);
		}
		
	
		DataViewBox.ShowAll();
		Application.Run();
	}
	
	///<summary>Displays a message in the status bar for 5 sec</summary>
	void DisplayStatusMessage(string message)
	{
		MessageStatusbar.Pop(0);	
		MessageStatusbar.Push(0, message);
		StatusMessageTimer.Change(5000, 0);	
	}
	
	///<summary>Clears the status bar message (callback)</summary>
	void ClearStatusMessage(object o)
	{
		MessageStatusbar.Pop(0);	
		//MessageStatusbar.Push(0, message);
		StatusMessageTimer.Change(System.Threading.Timeout.Infinite, System.Threading.Timeout.Infinite);	
	}
	
	///<summary>
	/// Returns the ByteBuffer's file path or "Untitled" if path is empty
	///</summary>
	string GetBufferPath(ByteBuffer bb)
	{
		if (bb.Filename==null || bb.Filename.Length == 0)
			return "Untitled";
		else
			return bb.Filename;
	}
	
	///<summary>
	/// Create and setup a DataView 
	///</summary>
	DataView CreateDataView(ByteBuffer bb)
	{
		DataView dv=new DataView();
		dv.SelectionChanged += UpdateSelectionStatusbar; //OnDataViewSelectionChanged;
		dv.CursorChanged += OnDataViewCursorChanged;
		dv.OverwriteChanged += UpdateOverwriteStatusbar; //OnDataViewOverwriteChanged;
	
		dv.Buffer=bb;
		
		return dv;
	}
	
	 
	///<summary>
	/// Updates the window's title with data from the specified DataView,
	/// if the specified DataView is active
	///</summary>
	void UpdateWindowTitle(DataView dv)
	{
		if (dataBook.NPages == 0)
			return;
		
		DataView curdv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		ByteBuffer curbb=curdv.Buffer;
		
		ByteBuffer bb=dv.Buffer;
		string filename=GetBufferPath(bb);
		
		// if DataView is active
		if (curbb==bb) {
			if (bb.HasChanged)
				MainWindow.Title= filename + " * - Bless";
			else
				MainWindow.Title= filename + " - Bless";
		}
	}
	
	///<summary>
	/// Updates the title of the tab with the specified DataView
	///</summary>
	void UpdateTabLabel(DataView dv)
	{
		ByteBuffer bb=dv.Buffer;
		string filename=GetBufferPath(bb);
		
		if (bb.HasChanged)
			dataBook.SetTabLabelText(dv.Display, Path.GetFileName(filename) +"*");
		else
			dataBook.SetTabLabelText(dv.Display, Path.GetFileName(filename));
	}
	
	///<summary>
	/// Updates the undo/redo items with data from the specified DataView,
	/// if the specified DataView is active
	///</summary>
	void UpdateUndoRedo(DataView dv)
	{
		if (dataBook.NPages == 0)
			return;
		
		DataView curdv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		ByteBuffer curbb=curdv.Buffer;
	
		ByteBuffer bb=dv.Buffer;
	
		// if DataView is active
		if (curbb==bb) {
			// gray Undo if there is no undo action
			// or we aren't allowed to modify the file
			if (bb.CanUndo && bb.ModifyAllowed)
				EditUndoMenuItem.Sensitive=true;
			else 
				EditUndoMenuItem.Sensitive=false;
	
			// gray Redo if there is no redo action
			// or we aren't allowed to modify the file
			if (bb.CanRedo && bb.ModifyAllowed)
				EditRedoMenuItem.Sensitive=true;
			else
				EditRedoMenuItem.Sensitive=false;
		}	
	
	}
	
	///<summary>
	/// Updates various menu items according to the FileOperationsAllowed
	/// permission property of the active ByteBuffer
	///</summary>
	void UpdateForFileOperationsAllowed(DataView dv)
	{
		if (dataBook.NPages == 0)
			return;
		
		DataView curdv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		ByteBuffer curbb=curdv.Buffer;
	
		ByteBuffer bb=dv.Buffer;
		
		// if DataView is active
		if (curbb==bb) {
			// gray various menu items
			// if FileOperations aren't allowed
			if (!bb.FileOperationsAllowed) {
				FileSaveMenuItem.Sensitive=false;
				FileSaveAsMenuItem.Sensitive=false;
				FileCloseMenuItem.Sensitive=false;
			}
			else {
				FileSaveMenuItem.Sensitive=true;
				FileSaveAsMenuItem.Sensitive=true;
				FileCloseMenuItem.Sensitive=true;
			}
		}
		
		
		// If even one ByteBuffer in the dataBook
		// doesn't allow file operations
		// gray the Quit menu item.
		FileQuitMenuItem.Sensitive=true; // set initially to true
		
		foreach (DataViewDisplay dd in dataBook.Children) {
			if (!dd.View.Buffer.FileOperationsAllowed) {
				FileQuitMenuItem.Sensitive=false;
				dataBook.SetCloseSensitivity(dd, false);
			}
			else
				dataBook.SetCloseSensitivity(dd, true);
		}
		
			
	}
	
	///<summary>
	/// Updates various menu items according to the ModifyAllowed 
	/// permission property of the active ByteBuffer
	///</summary>
	void UpdateForModifyAllowed(DataView dv)
	{
		if (dataBook.NPages == 0)
			return;
		
		DataView curdv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		ByteBuffer curbb=curdv.Buffer;
	
		ByteBuffer bb=dv.Buffer;
	
		// if DataView is active
		if (curbb==bb) {
			// gray various menu items
			// if we are not allowed to modify the buffer 
			if (!bb.ModifyAllowed) {
				EditCutMenuItem.Sensitive=false;
				EditPasteMenuItem.Sensitive=false;
				EditDeleteMenuItem.Sensitive=false;
				
			}
			else {
				EditCutMenuItem.Sensitive=true;
				EditPasteMenuItem.Sensitive=true;
				EditDeleteMenuItem.Sensitive=true;
			}
		}
	}
	
	///<summary>
	/// Updates various menu items and the corresponding DataView 
	/// according to the ReadAllowed permission property of the active ByteBuffer
	///</summary>
	void UpdateForReadAllowed(DataView dv)
	{
		if (dataBook.NPages == 0)
			return;
		
		DataView curdv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		ByteBuffer curbb=curdv.Buffer;
	
		ByteBuffer bb=dv.Buffer;
		
		// if DataView is active
		if (curbb==bb) {
			// gray various menu items and the whole dataview
			// if we are not allowed to read from the buffer  
			if (!bb.ReadAllowed) {
				EditCopyMenuItem.Sensitive=false;
				dv.Display.Sensitive=false;
			}
			else {
				EditCopyMenuItem.Sensitive=true;
				dv.Display.Sensitive=true;
			}
		}
	}
	
	///<summary>
	/// Updates the cursor status bar with data from the specified DataView,
	/// if the specified DataView is active
	///</summary>
	public void UpdateCursorStatusbar(DataView dv)
	{	
		if (dataBook.NPages == 0)
			return;
		
		DataView curdv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		if (curdv!=dv)
			return;
	
		long coffset=dv.CursorOffset;
		OffsetStatusbar.Pop(0);
		string str=string.Format("Offset: 0x{0:x} / 0x{1:x}", coffset, dv.Buffer.Size-1);
		OffsetStatusbar.Push(0, str);
	}
	
	///<summary>
	/// Updates the selection status bar with data from the specified DataView,
	/// if the specified DataView is active
	///</summary>
	public void UpdateSelectionStatusbar(DataView dv)
	{
		if (dataBook.NPages == 0)
			return;

		DataView curdv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		if (curdv!=dv)
			return;
	
		Bless.Util.Range sel=dv.Selection;
		sel.Sort();
		SelectionStatusbar.Pop(0);
		string str;
	
		if (sel.IsEmpty()==true)
			str="Selection: None";
		else	
			str=string.Format("Selection: 0x{0:x} to 0x{1:x} (0x{2:x} bytes)", sel.Start, sel.End, sel.Size);
	
		SelectionStatusbar.Push(0, str);
	}
	
	///<summary>
	/// Updates the overwrite status bar with data from the specified DataView,
	/// if the specified DataView is active
	///</summary>
	public void UpdateOverwriteStatusbar(DataView dv)
	{	
		if (dataBook.NPages == 0)
			return;
		
		DataView curdv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		if (curdv!=dv)
			return;
	
		OverwriteStatusbar.Pop(0);
		if (dv.Overwrite==true)
			OverwriteStatusbar.Push(0, " OVR");
		else
			OverwriteStatusbar.Push(0, " INS");
	}
	
	///<summary>
	/// Updates the File->Revert menu item with data from the specified DataView,
	/// if the specified DataView is active
	///</summary>
	public void UpdateRevert(DataView dv)
	{
		if (dataBook.NPages == 0)
			return;
		
		DataView curdv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		if (curdv!=dv)
			return;
			
		if (dv.Buffer.HasFile && dv.Buffer.ModifyAllowed)
			FileRevertMenuItem.Sensitive=true;
		else
			FileRevertMenuItem.Sensitive=false;
	}
	
	//
	// called the first time a pattern is sought
	//
	void OnFirstFind()
	{
		SearchFindNextMenuItem.Sensitive=true;
		SearchFindPreviousMenuItem.Sensitive=true;
	}
	
	//<summary>
	// Called when the cursor is moved
	//</summary>
	void OnDataViewCursorChanged(DataView dv)
	{
		if (dataBook.NPages == 0)
			return;
			
		DataView curdv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		if (curdv!=dv)
			return;
			
		UpdateCursorStatusbar(dv);
		conversionTable.Update(dv);
	}
	
	///<summary>Handle ByteBuffer changes</summary>
	void OnBufferChanged(ByteBuffer bb)
	{
		DataView dv=null;
		
		// find DataView that owns bb
		foreach (DataViewDisplay dvtemp in dataBook.Children) {
			if (dvtemp.View.Buffer==bb) {
				dv=dvtemp.View;
				break;	
			}
		}

		// if the status of bb has not changed since the
		// last invocation, there is no need to update
		// the titles
		if ((bool)bufferHasChanged[bb]!=bb.HasChanged) {
			bufferHasChanged[bb]=bb.HasChanged;
			UpdateWindowTitle(dv);
			UpdateTabLabel(dv);
		}
	
		// always update undo/redo and cursor status bar
		UpdateUndoRedo(dv);
		UpdateRevert(dv);
		UpdateForFileOperationsAllowed(dv);
		UpdateForModifyAllowed(dv);
		UpdateForReadAllowed(dv);
		OnDataViewCursorChanged(dv);
		
	}
	
	///<summary>Handle additions to history</summary>
	void OnHistoryChanged(History h)
	{
		Menu fileMenu=(Menu)FileMenuItem.Submenu;
		
		// clear file list
		while (fileMenu.Children.Length > 9)
			fileMenu.Remove(fileMenu.Children[7]);
		
		// add separator if it is needed
		if (h.Files.Count>1) {
			SeparatorMenuItem smi=new SeparatorMenuItem();
			fileMenu.Insert(smi, 7);
		}
		
		// add recent files
		int i=h.Files.Count;
		foreach (string file in h.Files) {
			HistoryMenuItem mi=new HistoryMenuItem(file, i);
			mi.Visible=true;
			mi.Activated+=OnHistoryMenuItemActivated;
			fileMenu.Insert(mi, 7);
			i--;
		}
		
		fileMenu.ShowAll();
	}
	
	///<summary>Handle Activated event on a HistoryMenuItem</summary>
	void OnHistoryMenuItemActivated(object o, EventArgs args)
	{
		HistoryMenuItem hmi=(HistoryMenuItem)o;
		
		// try to open the file
		ByteBuffer bb=OpenFile(hmi.FilePath);
		
		// if open was successful
		if (bb!=null) {
			if (dataBook.CanReplacePage(dataBook.CurrentPage)) { // replace current page
				DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
				dv.Buffer.CloseFile();
				dv.Buffer=bb;
				UpdateRevert(dv);
			}
			else { // create new page
				// create and setup a  DataView
				DataView dv=CreateDataView(bb);	
				dataBook.AppendView(dv, new CloseViewDelegate(CloseFile), Path.GetFileName(bb.Filename));	
			}
							
			OnBufferChanged(bb);
		}	
	}
	
	///<summary>If file has changed ask whether to save it.
	/// Returns true if file saved or the user doesn't want to save it.</summary>
	private bool AskForSaveIfFileChanged(DataView dv)
	{
		ByteBuffer bb=dv.Buffer;
	
		if (bb == null)
			return true;
	
		if (bb.HasChanged) {
			SaveConfirmationAlert sca=new SaveConfirmationAlert(bb.Filename, MainWindow);
			ResponseType res=(ResponseType)sca.Run();
			sca.Destroy();
	
			if (res== ResponseType.Ok) {
				return SaveFile(dv, null, false, true);
			}
			else if (res==ResponseType.No){
				return true;
			}
			else /*if (res==ResponseType.Cancel)*/
				return false;
		}
	
		return true;
	}
	
	///<summary>If any file has changed ask whether to save it.
	/// Returns true if all changed files are saved or the user doesn't want to save them.</summary>
	private bool AskForSaveIfFilesChanged()
	{
		ArrayList list=new ArrayList();
	
		// create a list with the changed files
		int i=0;
		foreach (DataViewDisplay dv in dataBook.Children) {
			ByteBuffer bb=dv.View.Buffer;
			if (bb.HasChanged)
			list.Add(new SaveFileItem(true, bb.Filename, i));
			i++;
		}
	
		if (list.Count == 0)
			return true;
	
		// special handling if only one file changed
		if (list.Count == 1) {
			int page=((SaveFileItem)list[0]).Page;
			// make the page active 
			dataBook.Page=page;
		
			DataView dv=((DataViewDisplay)dataBook.GetNthPage(page)).View;
			return AskForSaveIfFileChanged(dv);
		}
	
		// show the confirmation alert 
		SaveFileItem[] array=(SaveFileItem[])list.ToArray(typeof(SaveFileItem));
	
		SaveConfirmationMultiAlert sca=new SaveConfirmationMultiAlert(array, MainWindow);
		ResponseType res=(ResponseType)sca.Run();
		sca.Destroy();
	
		// handle responses	
		if (res == ResponseType.Ok) {
			bool allSaved=true;
			
			// save the files the user specified
			foreach(SaveFileItem item in array) {
				if (item.Save==true) {
					DataView dv=((DataViewDisplay)dataBook.GetNthPage(item.Page)).View;
					// make page active
					dataBook.Page=item.Page;
					// try to save the file
					if (SaveFile(dv, null, false, true)==false)
						allSaved=false;
					else {
						UpdateTabLabel(dv);
						UpdateWindowTitle(dv);	
					}	
				}
			} 
	
			return allSaved;
		}
		else if (res==ResponseType.No){
			return true;
		}
		else /*if (res==ResponseType.Cancel)*/
			return false;
	}
	
	
	///<summary>
	/// Try to open the file as a ByteBuffer 
	///</summary>
	private ByteBuffer OpenFile(string filename)
	{
		string fullPath=null;
	
		// get the full path
		try {
			fullPath=Path.GetFullPath(filename);	
		}
		catch(Exception ex) {
			ErrorAlert ea=new ErrorAlert("Error opening file "+filename, ex.Message, MainWindow);
			ea.Run();
			ea.Destroy();
		}
	
		// if we can't get full path return
		if (fullPath==null)
			return null;
	
		try {	
			ByteBuffer bb=ByteBuffer.FromFile(fullPath);
			bb.UseGLibIdle=true;
			DisplayStatusMessage("Loaded file '"+ fullPath +"'");
			bb.Changed+=OnBufferChanged;
			bufferHasChanged[bb]=true;
			history.Add(fullPath);
			return bb;
		}
		catch(UnauthorizedAccessException ex) {
			ErrorAlert ea=new ErrorAlert("Error opening file '"+ fullPath+"'", "You do not have read permissions for the file you requested.", MainWindow);
			ea.Run();
			ea.Destroy();
		}
		catch(System.IO.FileNotFoundException ex) {
			ErrorAlert ea=new ErrorAlert("Error opening file '"+ fullPath+"'", "The file you requested does not exist.", MainWindow);
			ea.Run();
			ea.Destroy();
		}
		catch(System.IO.IOException ex) {
			ErrorAlert ea=new ErrorAlert("Error opening file '"+ fullPath+"'", ex.Message, MainWindow);
			ea.Run();
			ea.Destroy();
		}
		catch(System.ArgumentException ex) {
			ErrorAlert ea=new ErrorAlert("Error opening file '"+ fullPath+"'", ex.Message, MainWindow);
			ea.Run();
			ea.Destroy();
		}
		
		return null;
	}
	
	///<summary>
	/// Manage high-level file saving procedures (eg confirm file overwrites, get filenames etc) 
	///</summary>
	private bool SaveFile(DataView dv, string filename, bool forceSaveAs, bool synchronous)
	{
		ByteBuffer bb=dv.Buffer;
	
	 	if (!dv.Buffer.FileOperationsAllowed) {
	 		return false;
	 	}
	 	
		// if a filename is given, save the file under the specified filename
		if (!forceSaveAs && filename!=null) 
			return SaveFileInternal(dv, filename, synchronous);
	  
		// if a filename is not given but buffer
		// has a file associated with it, save it under the same filename
		if (!forceSaveAs && filename==null && bb.HasFile==true && bb.Filename.Length!=0)
			return SaveFileInternal(dv, bb.Filename, synchronous);	  
	  
		// otherwise prompt user for a name	  
		Gtk.FileSelection fs=new Gtk.FileSelection("Save File As");
		bool done=false;
		bool fileSaved=true;
	
		do {
			ResponseType response = (ResponseType)fs.Run();
			fs.Hide();
			if (response == ResponseType.Ok) {
				// check to see whether file exists and prompt user to confirm
				if (File.Exists(fs.Filename)) {
					FileOverwriteAlert ea=new FileOverwriteAlert(fs.Filename, MainWindow);
					ResponseType response1 = (ResponseType)ea.Run();
					ea.Destroy();
					if (response1 == ResponseType.Ok) {
						fileSaved=SaveFileInternal(dv, fs.Filename, synchronous);
						done=true;
					}
					else
						done=false;
				}
				else{ // !File.Exists(fs.Filename)
					fileSaved=SaveFileInternal(dv, fs.Filename, synchronous);
					done=true;
				}
			}
			else { // response!=ResponseType.OK
				done=true;
				fileSaved=false;
			}
		} while(!done);
	
		fs.Destroy();	
		
		UpdateRevert(dv);
		
		return fileSaved; 
	}
	
	
	///<summary>
	/// Manage low-level file saving procedures 
	///</summary>
	private bool SaveFileInternal(DataView dv, string filename, bool synchronous)
	{
		ByteBuffer bb=dv.Buffer;
	
		string fullPath=null;
	
		// get the full path
		try {
			fullPath=Path.GetFullPath(filename);	
		}
		catch(Exception ex) {
			ErrorAlert ea=new ErrorAlert("Error saving file "+filename, ex.Message, MainWindow);
			ea.Run();
			ea.Destroy();
		}
	
		// if we can't get full path, return
		if (fullPath==null)
			return false;
		
		try {
			string msg;
			if (fullPath!=bb.Filename)
				msg="Saving file '"+ bb.Filename +"' as '"+fullPath+"'";
			else
				msg="Saving file '"+ bb.Filename+"'";
				
			DisplayStatusMessage(msg+"...");
			
			// create a ProgressDialog
			// if save is synchronous, make it modal
			ProgressDialog pd=new ProgressDialog(msg);
			pd.Modal=synchronous;
			
			IAsyncResult ar;
			
			// Decide whether to save or save as
			if (fullPath!=bb.Filename)
				ar=bb.BeginSaveAs(fullPath, new ProgressCallback(pd.Update), new AsyncCallback(SaveFileAsyncCallback));
			else
				ar=bb.BeginSave(new ProgressCallback(pd.Update), new AsyncCallback(SaveFileAsyncCallback));
				
			// if save is synchronous wait for save to finish
			if (synchronous) {
				// while waiting update the gui
				while (ar.AsyncWaitHandle.WaitOne(50, true)==false) {
					while( Application.EventsPending() ) {
      					Application.RunIteration();
      				}
				}
				// find out if save succeeded
				SaveOperation bbs=(SaveOperation)ar.AsyncState;
				if (bbs.Result!=SaveOperation.OperationResult.Finished)
					return false;
				// add to history
				history.Add(bbs.SavePath);
				return true;
			}
		}
		catch(IOException ex) {
			string file;
			if (fullPath!=bb.Filename)
				file=fullPath;
			else
				file=bb.Filename;
				
			ErrorAlert ea=new ErrorAlert("Error saving file '"+ file +"'", ex.Message, MainWindow);
			ea.Run();
			ea.Destroy();
			
			DisplayStatusMessage("The file '" + file + "' has NOT been saved");
		}
		
		return false;
	}
	
	///<summary>
	/// Callback to call when a save operation has finished
	///</summary>
	void SaveFileAsyncCallback(IAsyncResult ar)
	{
		SaveAsOperation bbs=(SaveAsOperation)ar.AsyncState;
		
		DataViewDisplay dvDisplay=null;
		
		// find out to which dataview the saved buffer belongs
		foreach (DataViewDisplay dd in dataBook.Children) {
			if (dd.View.Buffer==bbs.Buffer) {
				dvDisplay=dd;
				break;
			}
		}
		
		if (bbs.Result==SaveAsOperation.OperationResult.Finished) { // save went ok
			if (bbs.SavePath!=bbs.Buffer.Filename)
				DisplayStatusMessage("The file has been saved as '" + bbs.SavePath +"'");	
			else
				DisplayStatusMessage("The file '" + bbs.SavePath+ "' has been saved");
			
			// update local bb entry status
			bufferHasChanged[bbs.Buffer]=bbs.Buffer.HasChanged;
			// add to history
			history.Add(bbs.SavePath);
			
			UpdateTabLabel(dvDisplay.View);
			UpdateWindowTitle(dvDisplay.View);
			return;
		}
		else if (bbs.Result==SaveAsOperation.OperationResult.Cancelled) { // save cancelled
		
		}
		else if (bbs.Result==SaveAsOperation.OperationResult.CaughtException) {
			// * UnauthorizedAccessException
			// * System.ArgumentException
			// * System.IO.IOException
			ErrorAlert ea=new ErrorAlert("Error saving file '"+ bbs.SavePath+"'", bbs.ThreadException.Message, MainWindow);
			ea.Run();
			ea.Destroy();
		}
		
		DisplayStatusMessage("The file '" + bbs.SavePath + "' has NOT been saved");
	}
	
	void CloseFile(DataView dv)
	{
		if (!dv.Buffer.FileOperationsAllowed)
	 		return;
	 	
		// if user decides not to close after all
		if (AskForSaveIfFileChanged(dv)==false)
			return;
	
		dv.Buffer.CloseFile();
		dataBook.RemovePage(dataBook.PageNum(dv.Display));
	
		// promptly free pixmap memory
		dv.Display.DisposePixmaps();
		dv=null;
		
		// Make sure the new current DataView has the focus
		DataViewDisplay dvd=(DataViewDisplay)dataBook.CurrentPageWidget;
		if (dvd!=null)
			dvd.GrabKeyboardFocus();
		else { 
			// there is no dataview left,
			// update title and statusbars
			MainWindow.Title="Bless - Gtk# Hex Editor";
			OffsetStatusbar.Pop(0);
			OffsetStatusbar.Push(0, "Offset: -");
			SelectionStatusbar.Pop(0);
			SelectionStatusbar.Push(0, "Selection: None");
			FileRevertMenuItem.Sensitive=false;
			conversionTable.Clear();
		}
	}
	
	///<summary>
	/// Try to quit. Ask user to save files before quitting and
	/// save the current session.
	///</summary>
	void TryQuit()
	{
	 	if (AskForSaveIfFilesChanged()) {
	 		string blessConfDir=Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), "bless");
			try {
				SaveSession(Path.Combine(blessConfDir,"last.session"));
				SaveHistory(Path.Combine(blessConfDir,"history.xml"));
			}
			catch (Exception ex) { }
			Application.Quit ();
		}
	}
	
	///<summary>
	/// Load the files specified
	///</summary>
	void LoadFiles(string[] files)
	{
		foreach(string file in files) {
			ByteBuffer bb=OpenFile(file);
			// if file was opened successfully
			if (bb!=null) {
				DataView dv=CreateDataView(bb);
				dataBook.AppendView(dv, new CloseViewDelegate(CloseFile), Path.GetFileName(bb.Filename));
				OnBufferChanged(bb);
			}
		}  
	}
	
	///<summary>
	/// Save the current session to the specified file.
	///</summary>
	void SaveSession(string path)
	{
		Session session=new Session();
		Gdk.Rectangle alloc=MainWindow.Allocation;
		
		// save the window size
		session.WindowHeight=alloc.Height;
		session.WindowWidth=alloc.Width;
		
		// save information about the currently open files
		foreach(DataViewDisplay dvd in dataBook.Children) {
			DataView dv=dvd.View;
			// add to session only if buffer is related to a file
			if (dv.Buffer.HasFile) {
				session.AddFile(dv.Buffer.Filename, dv.Offset, dv.CursorOffset, dv.CursorDigit, dvd.Layout.FilePath, 0);
				if (dv == ((DataViewDisplay)dataBook.CurrentPageWidget).View) {
					session.ActiveFile=dv.Buffer.Filename;
				}
			}	
		}
		session.Save(path);
	}
	
	///<summary>
	/// Load a session from the specified file.
	///</summary>
	void LoadSession(string path)
	{
		Session session=new Session();
		
		// try to load session file
		try {
			session.Load(path);
		}
		catch(Exception ex) {
			return;
		}
		
		// set the size of main window
		MainWindow.Resize(session.WindowWidth, session.WindowHeight);
		
		// add files to the DataBook
		foreach(SessionFileInfo sfi in session.Files) {
			ByteBuffer bb=OpenFile(sfi.Path);
			// if file was opened successfully
			if (bb!=null) {
				DataView dv=CreateDataView(bb);
				// try to load layout file
				try {
					dv.Display.Layout=new Bless.Gui.Layout(sfi.Layout);
				}
				catch(Exception ex) {
					ErrorAlert ea=new ErrorAlert("Error loading layout '"+ sfi.Layout +"' for file '"+sfi.Path+"'. Loading default layout.", ex.Message , MainWindow);
					ea.Run();
					ea.Destroy();
				}
				
				long cursorOffset=sfi.CursorOffset;
				
				// sanity check cursor offset and view offset
				if (cursorOffset > bb.Size) 
					cursorOffset = bb.Size;	
				dv.MoveCursor(cursorOffset, sfi.CursorDigit);
				
				long offset=sfi.Offset;
				if (offset >= bb.Size)
					offset = 0;
				dv.Offset=offset;
				
				dataBook.AppendView(dv, new CloseViewDelegate(CloseFile), Path.GetFileName(bb.Filename));
				OnBufferChanged(bb);
			}
		}
		
		foreach(DataViewDisplay dvd in dataBook.Children) {
			DataView dv=dvd.View;
			if (dv.Buffer.Filename == session.ActiveFile) {
				dataBook.CurrentPage=dataBook.PageNum(dvd);
				break;
			}
		}
	}
	
	///<summary>
	/// Save the history to the specified file
	///</summary>
	void SaveHistory(string path)
	{
		try {
			history.Save(path);
		}
		catch (Exception ex) {
		
		}
	}
	
	///<summary>
	/// Load the history from the specified file
	///</summary>
	void LoadHistory(string path)
	{
		try {
			history.Load(path);
		}
		catch (Exception ex) {
		
		}
	}
	
	///<summary>
	/// Callback for SwitchPage DataBook event 
	///</summary>
	void OnSwitchPage(object o, SwitchPageArgs args)
	{
		DataView dv=((DataViewDisplay)dataBook.GetNthPage((int)args.PageNum)).View;
	
		UpdateWindowTitle(dv);
		UpdateUndoRedo(dv);
		UpdateRevert(dv);
		UpdateForFileOperationsAllowed(dv);
		UpdateForModifyAllowed(dv);
		UpdateForReadAllowed(dv);
			
		UpdateCursorStatusbar(dv);
		UpdateSelectionStatusbar(dv);
		UpdateOverwriteStatusbar(dv);
		conversionTable.Update(dv);
	
		dv.Display.GrabKeyboardFocus();
	}
	
	///<summary>Assign the resize grip to a visible statusbar</summary>
	void AssignResizeGrip()
	{
		// clear all grips
		OffsetStatusbar.HasResizeGrip=false;
		SelectionStatusbar.HasResizeGrip=false;
		MessageStatusbar.HasResizeGrip=false;
		OverwriteStatusbar.HasResizeGrip=false;
		DummyStatusbar.Visible=false;
		
		// give resize grip to a statusbar
		if (OverwriteStatusbar.Visible==false){
			if (SelectionStatusbar.Visible==false) {
				if (OffsetStatusbar.Visible==false)
					DummyStatusbar.Visible=true; 
				else
					OffsetStatusbar.HasResizeGrip=true;
			}
			else
				SelectionStatusbar.HasResizeGrip=true;	
		}
		else
			OverwriteStatusbar.HasResizeGrip=true;
	}
	
	///<summary>Handle file->new command from menu</summary>
	public void OnFileNew(object o, EventArgs args) 
	{
		ByteBuffer bb=new ByteBuffer();
		bb.UseGLibIdle=true;
		bb.Changed+=OnBufferChanged;
		bufferHasChanged[bb]=true;
	
		// create and setup a  DataView
		DataView dv=CreateDataView(bb);
	
		dataBook.AppendView(dv, new CloseViewDelegate(CloseFile), Path.GetFileName(bb.Filename));
		OnBufferChanged(bb);
	}
	
	///<summary>Handle file->open command from menu</summary>
	public void OnFileOpen(object o, EventArgs args) 
	{
		// should we replace the current page or create a new one?
		bool replaceCurrentPage=dataBook.CanReplacePage(dataBook.CurrentPage);
		
		// Get path of file to open	
		Gtk.FileSelection fs=new Gtk.FileSelection("Open File");
		
		ResponseType result = (ResponseType)fs.Run();
		fs.Hide();
		
		// if user does not cancel
		if (result == ResponseType.Ok) {
			// try to open the file
			ByteBuffer bb=OpenFile(fs.Filename);

			// if open was successful
			if (bb!=null) {
				if (replaceCurrentPage) { // replace current page
					DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
					dv.Buffer.CloseFile();
					dv.Buffer=bb;
					UpdateRevert(dv);
				}
				else { // create new page
					// create and setup a  DataView
					DataView dv=CreateDataView(bb);	
					dataBook.AppendView(dv, new CloseViewDelegate(CloseFile), Path.GetFileName(bb.Filename));	
				}
				
				OnBufferChanged(bb);
			}	
		}
	
		fs.Destroy();
	}
	
	///<summary>Handle file->save command from menu</summary>
	public void OnFileSave(object o, EventArgs args)
	{
		if (dataBook.NPages == 0)
			return;
	
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
	
		SaveFile(dv, null, false, false);
	}
	
	///<summary>Handle file->save as command from menu</summary>
	public void OnFileSaveAs(object o, EventArgs args)
	{
		if (dataBook.NPages == 0)
			return;
	
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
	
		SaveFile(dv, null, true, false);
	}
	
	///<summary>Handle file->revert command from menu</summary>
	public void OnFileRevert(object o, EventArgs args)
	{
		if (dataBook.NPages == 0)
			return;
	
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		
		RevertConfirmationAlert rca=new RevertConfirmationAlert(dv.Buffer.Filename, MainWindow);
		ResponseType response=(ResponseType)rca.Run();
		rca.Destroy();
		
		try {
			if (response==ResponseType.Ok)
				dv.Revert();
		}
		catch (FileNotFoundException ex) {
			ErrorAlert ea=new ErrorAlert("Error reverting file '"+ ex.Message+"'.","The file cannot be found. Perhaps it has been recently deleted." , MainWindow);
			ea.Run();
			ea.Destroy();
		}
	}
	
	///<summary>Handle file->close as command from menu</summary>
	public void OnFileClose(object o, EventArgs args)
	{
		if (dataBook.NPages == 0)
			return;
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
	
		CloseFile(dv);
	}
	
	
	///<summary>Handle file->quit command from menu</summary>
	public void OnFileQuit(object o, EventArgs args) 
	{
		TryQuit();
	}
	
	///<summary>Handle edit->undo command from menu</summary>
	public void OnEditUndo(object o, EventArgs args) 
	{
		if (dataBook.NPages == 0)
			return;
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		dv.Undo();
	}
	
	///<summary>Handle edit->redo command from menu</summary>
	public void OnEditRedo(object o, EventArgs args) 
	{
		if (dataBook.NPages == 0)
			return;
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		dv.Redo();
	}
	
	///<summary>Handle edit->cut command from menu</summary>
	public void OnEditCut(object o, EventArgs args) 
	{
		if (dataBook.NPages == 0)
			return;
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		dv.Cut();
	}
	
	///<summary>Handle edit->copy command from menu</summary>
	public void OnEditCopy(object o, EventArgs args) 
	{
		if (dataBook.NPages == 0)
			return;
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		dv.Copy();
	}
	
	///<summary>Handle edit->paste command from menu</summary>
	public void OnEditPaste(object o, EventArgs args) 
	{
		if (dataBook.NPages == 0)
			return;
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		dv.Paste();
	}
	
	///<summary>Handle edit->delete command from menu</summary>
	public void OnEditDelete(object o, EventArgs args) 
	{
		if (dataBook.NPages == 0)
			return;
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		dv.Delete();
	}
	
	///<summary>Handle the View->Layouts command</summary>
	public void OnViewLayouts(object o, EventArgs args)
	{
		// show the selection dialog
		new LayoutSelectionDialog(dataBook);
	}
	
	///<summary>Handle the View->Statusbar->Show command</summary>
	public void OnViewStatusbarShow(object o, EventArgs args)
	{
		if (ViewStatusbarShowMenuItem.Active) {
			StatusPaned.Visible=true;
			ViewStatusbarOffsetMenuItem.Sensitive=true;
			ViewStatusbarSelectionMenuItem.Sensitive=true;
			ViewStatusbarOverwriteMenuItem.Sensitive=true;
		}
		else {
			StatusPaned.Visible=false;
			ViewStatusbarOffsetMenuItem.Sensitive=false;
			ViewStatusbarSelectionMenuItem.Sensitive=false;
			ViewStatusbarOverwriteMenuItem.Sensitive=false;
		}
	}
	
	///<summary>Handle the View->Statusbar->Offset command</summary>
	public void OnViewStatusbarOffset(object o, EventArgs args)
	{
		if (ViewStatusbarOffsetMenuItem.Active)
			OffsetStatusbar.Visible=true;
		else
			OffsetStatusbar.Visible=false;
			
		AssignResizeGrip();
	}
	
	///<summary>Handle the View->Statusbar->Selection command</summary>
	public void OnViewStatusbarSelection(object o, EventArgs args)
	{
		if (ViewStatusbarSelectionMenuItem.Active)
			SelectionStatusbar.Visible=true;
		else
			SelectionStatusbar.Visible=false;
		
		AssignResizeGrip();
	}
	
	///<summary>Handle the View->Statusbar->Selection command</summary>
	public void OnViewStatusbarOverwrite(object o, EventArgs args)
	{
		if (ViewStatusbarOverwriteMenuItem.Active)
			OverwriteStatusbar.Visible=true;
		else
			OverwriteStatusbar.Visible=false;
		
		AssignResizeGrip();
	}
	
	///<summary>Handle the View->Statusbar->Selection command</summary>
	public void OnViewConversionTable(object o, EventArgs args)
	{
		if (ViewConversionTableMenuItem.Active)
			conversionTable.Visible=true;
		else
			conversionTable.Visible=false;
	}
	
	///<summary>Handle the Search->Find command</summary>
	public void OnSearchFind(object o, EventArgs args)
	{
		// get pattern from current selection...
		if (dataBook.NPages > 0) {
			DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
			findDialog.LoadWithSelection(dv);
		}
		
		findDialog.Show();
	}
	
	///<summary>Handle the Search->Find Next command</summary>
	public void OnSearchFindNext(object o, EventArgs args)
	{
		finder.FindNext(FindNextAsyncCallback);
	}
	
	void FindNextAsyncCallback(IAsyncResult ar)
	{
		ThreadedAsyncResult tar=(ThreadedAsyncResult) ar;
		
		FindNextOperation state=(FindNextOperation)tar.AsyncState;
		if (state.Result==FindNextOperation.OperationResult.Finished && state.Match==null) {
			InformationAlert ia=new InformationAlert("The pattern you requested was not found.", "End of file reached.", MainWindow);
			ia.Run();
			ia.Destroy();
		}
	}
	
	///<summary>Handle the Search->Find Previous command</summary>
	public void OnSearchFindPrevious(object o, EventArgs args)
	{
		finder.FindPrevious(FindPreviousAsyncCallback);
	}
	
	void FindPreviousAsyncCallback(IAsyncResult ar)
	{
		ThreadedAsyncResult tar=(ThreadedAsyncResult) ar;
		
		FindPreviousOperation state=(FindPreviousOperation)tar.AsyncState;
		if (state.Result==FindPreviousOperation.OperationResult.Finished && state.Match==null) {
			InformationAlert ia=new InformationAlert("The pattern you requested was not found.", "Beginning of file reached.", MainWindow);
			ia.Run();
			ia.Destroy();
		}
	}
	
	///<summary>Handle the Search->Find Previous command</summary>
	public void OnSearchReplace(object o, EventArgs args)
	{
		// get search pattern from current selection...
		if (dataBook.NPages > 0) {
			DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
			replaceDialog.LoadWithSelection(dv);
		}
		
		replaceDialog.Show();
	}
	
	///<summary>Handle the Search->Find Next command</summary>
	public void OnSearchGotoOffset(object o, EventArgs args)
	{
		gotoOffsetDialog.Show();
	}
	
	///<summary>Handle the Help->Contents command</summary>
	public void OnHelpContents(object o, EventArgs args)
	{
#if ENABLE_UNIX_SPECIFIC
		string helpScript=FileResourcePath.GetPath("..", "data", "help_script.sh");
		System.Diagnostics.Process.Start(helpScript);
#endif
	}
	
	///<summary>Handle the Help->About command</summary>
	public void OnHelpAbout(object o, EventArgs args)
	{
		AboutDialog.Show();
	}
	
	///<summary>Handle Main Window Delete Event</summary>
	public void OnMainWindowDeleteEvent (object o, DeleteEventArgs args) 
	{
		// make sure we can quit safely:
		// 1. no file operations are currently going on (eg save, replace etc)
		// 2. user has saved or doesn't want to save their modified files
		if (FileQuitMenuItem.Sensitive)
			TryQuit();
		args.RetVal = true;
	}

} // end bless main

